import requests
from uuid import uuid4
import json
import time
from typing import Sequence, Generator
from tqdm import tqdm


class SOCOClient(object):
    def __init__(self, api_key, host=None):
        self.api_key = api_key
        if host is None:
            self._server_url = 'https://api.soco.ai'
        else:
            self._server_url = host

        # MONITOR
        self.status_url = self._server_url + '/v1/index/status'

        # QUERY
        self.query_url = self._server_url + '/v1/search/query'

        # INDEX
        self.publish_url = self._server_url + '/v1/index/publish'
        self.abort_url = self._server_url + '/v1/index/abort'
        self.add_url = self._server_url + '/v1/index/add'
        self.read_url = self._server_url + '/v1/index/read'
        self.delete_url = self._server_url + '/v1/index/delete'

    def _get_header(self):
        return {'Content-Type': 'application/json', "Authorization": self.api_key}

    def _check_fields(self, fields, object):
        for f in fields:
            if f not in object['answer']:
                raise Exception("{} is required inside answer".format(f))

    def _check_frame_format(self, data):
        for frame in data:
            if 'answer' not in frame:
                raise Exception("answer is required for frame")
            self._check_fields(['value', 'context', 'answer_start'], frame)

            # check questions
            if 'questions' in frame:
                if type(frame['questions']) is not list:
                    raise Exception("Expect list for questions")
                for q in frame['questions']:
                    self._check_fields(['value'], q)

    def _check_doc_format(self, data):
        for doc in data:
            if 'data' not in doc:
                raise Exception("data is required for a doc")

            if type(doc['data']) is not list:
                raise Exception("data should be a list.")

            # check questions
            if 'meta' in doc:
                if type(doc['meta']) is not dict:
                    raise Exception("meta should be a dict")

    def _chunks(self, l: Sequence, n: int = 100) -> Generator[Sequence, None, None]:
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(l), n):
            is_last = i + n >= len(l)
            is_first = i == 0
            yield l[i:i + n]

    def wait_for_ready(self, check_frequency=15, timeout=-1, verbose=False):
        start_time = time.time()
        time.sleep(0.5)
        while True:
            state = self.status()
            if state['status'] == 'ready':
                break

            elapsed_time = int(time.time() - start_time)
            if verbose:
                print("{} seconds elapsed. status={}. {}/{} indexed. {} frames completed.".format(
                    int(time.time() - start_time),
                    state['status'],
                    state['publish_progress'],
                    state['num_documents'],
                    state['size']))

            if 0 < timeout < elapsed_time:
                print("Time out!")
                return

            time.sleep(check_frequency)

    def query(self, query, aggs=None, uid=None):
        data = {
            "query": query,
            "aggs": aggs,
            "uid": uid if uid is not None else str(uuid4())
        }
        result = requests.post(self.query_url, json=data, headers=self._get_header(), timeout=60)
        if result.status_code >= 300:
            raise Exception("Error in connecting to the SOCO servers")

        return json.loads(result.text)

    def status(self):
        result = requests.get(self.status_url, headers=self._get_header())
        if result.status_code >= 300:
            raise Exception("Error in connecting to the SOCO servers")

        return json.loads(result.text)

    @classmethod
    def pprint(cls, results):
        for r in results['results']:
            print("({}) - {}".format(r['score'], r['a']['value']))

    def add_data(self, data):
        self._check_doc_format(data)
        job_results = []
        batch_size = 50
        for batch in tqdm(self._chunks(data, n=batch_size),
                          desc='Uploading {} docs to task'.format(len(data)),
                          total=len(data) / batch_size):
            body = {"data": batch}
            result = requests.post(self.add_url, json=body, headers=self._get_header())
            if result.status_code >= 300:
                raise Exception("Error in appending to index at SOCO servers")

            job_results.append(json.loads(result.text))

        return job_results

    def read_data(self, batch_size=100):
        data = []
        skip = 0
        limit = batch_size
        while True:
            results = requests.post(self.read_url, json={'skip': skip, 'limit': limit}, headers=self._get_header())
            batch_docs = results.json()
            data.extend(batch_docs)
            if len(batch_docs) < limit:
                break
            skip = len(data)

        return data

    def delete_data(self, doc_ids=None):
        if doc_ids is None:
            print("Delete all data")
            result = requests.post(self.delete_url, headers=self._get_header())
        else:
            result = requests.post(self.delete_url,
                                   json={'doc_ids': doc_ids},
                                   headers=self._get_header())
        return result

    def publish(self, encoder_id=None, db_encoder_id=None, publish_args=None, sync=True):
        body = {}
        if encoder_id is not None:
            body['encoder_id'] = encoder_id

        if db_encoder_id is not None:
            body['db_encoder_id'] = db_encoder_id

        if publish_args is not None:
            body['publish_args'] = publish_args

        result = requests.post(self.publish_url, json=body, headers=self._get_header())
        if result.status_code > 299:
            raise Exception(result.json())
        if sync:
            self.wait_for_ready(verbose=True)
            print("Index is ready!")

        return result

    def abort(self, sync=True):
        result = requests.post(self.abort_url, headers=self._get_header())
        if sync:
            self.wait_for_ready(verbose=True)

        return result



import re


def split_sentence(sentence, lang='en'):
    if lang == 'en':
        return en_split_sentence(sentence)
    elif lang == 'zh':
        return zh_split_sentence(sentence)
    else:
        raise Exception("Unsupported language {}".format(lang))


def zh_split_sentence(text):
    resentencesp = re.compile('([\t﹒﹔﹖﹗．；。！？]["’”」』]{0,2}|：(?=["‘“「『]{1,2}|$))')

    s = text
    slist = []
    for i in resentencesp.split(s):
        if resentencesp.match(i) and slist:
            slist[-1] += i
        elif i:
            slist.append(i)
    return slist


def en_split_sentence(text):
    sentences = []
    for span in _infer_sentence_breaks(text):
        sentences.append(text[span[0]:span[1]])

    return sentences


def _infer_sentence_breaks(uni_text):
    """Generates (start, end) pairs demarking sentences in the text.

  Args:
    uni_text: A (multi-sentence) passage of text, in Unicode.

  Yields:
    (start, end) tuples that demarcate sentences in the input text. Normal
    Python slicing applies: the start index points at the first character of
    the sentence, and the end index is one past the last character of the
    sentence.
  """
    # Treat the text as a single line that starts out with no internal newline
    # characters and, after regexp-governed substitutions, contains internal
    # newlines representing cuts between sentences.
    uni_text = re.sub(r'\n', r' ', uni_text)  # Remove pre-existing newlines.
    text_with_breaks = _sed_do_sentence_breaks(uni_text)
    starts = [m.end() for m in re.finditer(r'^\s*', text_with_breaks, re.M)]
    sentences = [s.strip() for s in text_with_breaks.split('\n')]
    assert len(starts) == len(sentences)
    for i in range(len(sentences)):
        start = starts[i]
        end = start + len(sentences[i])
        yield start, end


def _sed_do_sentence_breaks(uni_text):
    """Uses regexp substitution rules to insert newlines as sentence breaks.

  Args:
    uni_text: A (multi-sentence) passage of text, in Unicode.

  Returns:
    A Unicode string with internal newlines representing the inferred sentence
    breaks.
  """

    # The main split, looks for sequence of:
    #   - sentence-ending punctuation: [.?!]
    #   - optional quotes, parens, spaces: [)'" \u201D]*
    #   - whitespace: \s
    #   - optional whitespace: \s*
    #   - optional opening quotes, bracket, paren: [['"(\u201C]?
    #   - upper case letter or digit
    txt = re.sub(r'''([.?!][)'" %s]*)\s(\s*[['"(%s]?[A-Z0-9])''' % ('\u201D', '\u201C'),
                 r'\1\n\2',
                 uni_text)

    # Wiki-specific split, for sentence-final editorial scraps (which can stack):
    #  - ".[citation needed]", ".[note 1] ", ".[c] ", ".[n 8] "
    txt = re.sub(r'''([.?!]['"]?)((\[[a-zA-Z0-9 ?]+\])+)\s(\s*['"(]?[A-Z0-9])''',
                 r'\1\2\n\4', txt)

    # Wiki-specific split, for ellipses in multi-sentence quotes:
    # "need such things [...] But"
    txt = re.sub(r'(\[\.\.\.\]\s*)\s(\[?[A-Z])', r'\1\n\2', txt)

    # Rejoin for:
    #   - social, military, religious, and professional titles
    #   - common literary abbreviations
    #   - month name abbreviations
    #   - geographical abbreviations
    #
    txt = re.sub(r'\b(Mrs?|Ms|Dr|Prof|Fr|Rev|Msgr|Sta?)\.\n', r'\1. ', txt)
    txt = re.sub(r'\b(Lt|Gen|Col|Maj|Adm|Capt|Sgt|Rep|Gov|Sen|Pres)\.\n',
                 r'\1. ',
                 txt)
    txt = re.sub(r'\b(e\.g|i\.?e|vs?|pp?|cf|a\.k\.a|approx|app|es[pt]|tr)\.\n',
                 r'\1. ',
                 txt)
    txt = re.sub(r'\b(Jan|Aug|Oct|Nov|Dec)\.\n', r'\1. ', txt)
    txt = re.sub(r'\b(Mt|Ft)\.\n', r'\1. ', txt)
    txt = re.sub(r'\b([ap]\.m)\.\n(Eastern|EST)\b', r'\1. \2', txt)

    # Rejoin for personal names with 3,2, or 1 initials preceding the last name.
    txt = re.sub(r'\b([A-Z]\.)[ \n]([A-Z]\.)[ \n]([A-Z]\.)[ \n]("?[A-Z][a-z])',
                 r'\1 \2 \3 \4',
                 txt)
    txt = re.sub(r'\b([A-Z]\.)[ \n]([A-Z]\.)[ \n]("?[A-Z][a-z])',
                 r'\1 \2 \3',
                 txt)
    txt = re.sub(r'\b([A-Z]\.[A-Z]\.)\n("?[A-Z][a-z])', r'\1 \2', txt)
    txt = re.sub(r'\b([A-Z]\.)\n("?[A-Z][a-z])', r'\1 \2', txt)

    # Resplit for common sentence starts:
    #   - The, This, That, ...
    #   - Meanwhile, However,
    #   - In, On, By, During, After, ...
    txt = re.sub(r'([.!?][\'")]*) (The|This|That|These|It) ', r'\1\n\2 ', txt)
    txt = re.sub(r'(\.) (Meanwhile|However)', r'\1\n\2', txt)
    txt = re.sub(r'(\.) (In|On|By|During|After|Under|Although|Yet|As |Several'
                 r'|According to) ',
                 r'\1\n\2 ',
                 txt)

    # Rejoin for:
    #   - numbered parts of documents.
    #   - born, died, ruled, circa, flourished ...
    #   - et al (2005), ...
    #   - H.R. 2000
    txt = re.sub(r'\b([Aa]rt|[Nn]o|Opp?|ch|Sec|cl|Rec|Ecl|Cor|Lk|Jn|Vol)\.\n'
                 r'([0-9IVX]+)\b',
                 r'\1. \2',
                 txt)
    txt = re.sub(r'\b([bdrc]|ca|fl)\.\n([A-Z0-9])', r'\1. \2', txt)
    txt = re.sub(r'\b(et al)\.\n(\(?[0-9]{4}\b)', r'\1. \2', txt)
    txt = re.sub(r'\b(H\.R\.)\n([0-9])', r'\1 \2', txt)

    # SQuAD-specific joins.
    txt = re.sub(r'(I Am\.\.\.)\n(Sasha Fierce|World Tour)', r'\1 \2', txt)
    txt = re.sub(r'(Warner Bros\.)\n(Records|Entertainment)', r'\1 \2', txt)
    txt = re.sub(r'(U\.S\.)\n(\(?\d\d+)', r'\1 \2', txt)
    txt = re.sub(r'\b(Rs\.)\n(\d)', r'\1 \2', txt)

    # SQuAD-specific splits.
    txt = re.sub(r'\b(Jay Z\.) ([A-Z])', r'\1\n\2', txt)
    txt = re.sub(r'\b(Washington, D\.C\.) ([A-Z])', r'\1\n\2', txt)
    txt = re.sub(r'\b(for 4\.\)) ([A-Z])', r'\1\n\2', txt)
    txt = re.sub(r'\b(Wii U\.) ([A-Z])', r'\1\n\2', txt)
    txt = re.sub(r'\. (iPod|iTunes)', r'.\n\1', txt)
    txt = re.sub(r' (\[\.\.\.\]\n)', r'\n\1', txt)
    txt = re.sub(r'(\.Sc\.)\n', r'\1 ', txt)
    txt = re.sub(r' (%s [A-Z])' % '\u2022', r'\n\1', txt)
    return txt



from soco_core import sentence_splitter
from uuid import uuid4
import re


class DocConvert(object):
    @classmethod
    def _get_context(cls, data, current_idx, prev_char, next_char, last_title, last_section):
        answer = data[current_idx]['text']
        budget = prev_char
        idx = current_idx - 1
        prev_context = []

        # COMPUTE PREV CONTEXT
        while idx >= 0:
            if budget <= 0:
                break

            text = data[idx]['text']
            if data[idx]['type'] in ['title', 'section']:
                break

            if text not in prev_context:
                prev_context.append(text)
                budget = budget - len(text)

            idx -= 1

        if last_section and last_section != answer:
            prev_context.append(last_section)

        if last_title and last_title != answer:
            prev_context.append(last_title)

        prev_context = prev_context[::-1]

        # COMPUTE NEXT CONTEXT
        budget = next_char
        idx = current_idx + 1
        next_context = []
        while idx < len(data):
            if budget <= 0:
                break
            if data[idx]['type'] == 'title':
                break
            next_context.append(data[idx]['text'])
            budget = budget - len(data[idx]['text'])
            idx += 1

        context = ' '.join(prev_context + [answer] + next_context)
        answer_start = len(' '.join(prev_context)) + 1 if len(prev_context) > 0 else 0
        assert context[answer_start:answer_start + len(answer)] == answer
        return {'context': context,
                'answer_start': answer_start,
                'value': answer,
                'displayable': False,
                'add_to_qa_index': True}

    @classmethod
    def _normalize_whitespace(cls, text):
        text= text.replace('\n', ' ').replace('\r', '').replace('\t', '').replace('\xa0', ' ')
        return text

    @classmethod
    def context2frame(cls, context, meta):
        frame = {'answer': context, 'meta': meta}
        return frame

    @classmethod
    def get_context_size(cls, lang):
        if lang == 'en':
            return 150
        elif lang == 'zh':
            return 75
        else:
            raise Exception("unknown language {}".format(lang))

    @classmethod
    def document_to_frames(cls, doc, lang, doc_meta=None, verbose=False, min_ans_len=5, max_ans_len=500):
        # CUT DOCUMENTS INTO SENTENCES
        flatten_data = []
        too_short_cnt = 0
        too_long_cnt = 0

        for chunk in doc:
            text = chunk['text']
            if text is None:
                continue

            text = cls._normalize_whitespace(text)
            uid = str(uuid4())
            chunk_type = chunk.get('type')

            if chunk_type != 'content':
                record = {'text': text, 'chunk_id': uid, 'type': chunk_type, 'answer_start': 0}
                flatten_data.append(record)
            else:
                sentences = sentence_splitter.split_sentence(text, lang=lang)
                for s in sentences:
                    if len(s) < min_ans_len:
                        too_short_cnt += 1
                        continue

                    if len(s) > max_ans_len:
                        too_long_cnt += 1
                        continue

                    answer_start = text.index(s)
                    record = {'text': s, 'chunk_id': uid, 'type': chunk_type, 'answer_start': answer_start}
                    flatten_data.append(record)
                    assert text[answer_start:answer_start + len(s)] == s

        # INDEX RAW SENTENCES AS FRAMES
        last_title = ''
        last_section = ''
        frames = []
        for f_id, f_data in enumerate(flatten_data):
            chunk_type = f_data['type']
            text = f_data['text']
            if chunk_type == 'title':
                last_title = text
            elif chunk_type == 'section':
                last_section = text

            context_size = cls.get_context_size(lang)
            if chunk_type in ['title', 'section']:
                context = cls._get_context(flatten_data, f_id, prev_char=0, next_char=context_size,
                                           last_title=last_title, last_section=last_section)
            else:
                context = cls._get_context(flatten_data, f_id, prev_char=context_size, next_char=context_size,
                                           last_title=last_title, last_section=last_section)

            meta = {'chunk_id': f_data['chunk_id'], 'chunk_type': chunk_type}
            if doc_meta is not None:
                meta.update(**doc_meta)

            frame = cls.context2frame(context, meta)

            frames.append(frame)

        if verbose:
            print("DONE PROCESS {} RAW DOCUMENTS with {} too short skip {} too long skip".format(len(frames),
                                                                                                 too_short_cnt,
                                                                                                 too_long_cnt))
        return frames



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = '0.1.4'
setuptools.setup(
    name="soco-core-python",
    version=VERSION,
    author="tinachez",
    description="Python SDK for using SOCO platform",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.soco.ai",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
    ],
    install_requires = [
        'nltk >= 3.4',
        'tqdm >= 4.32.1'
    ]
)


import json
from soco_core.convertors import DocConvert
import os


def load_example_frame_data(name):
    cur_path = os.path.dirname(os.path.abspath(__file__))
    if name == 'mr-sun':
        doc = json.load(open(os.path.join(cur_path, '..', 'resources/mr-sun.json')))
        frames = DocConvert.document_to_frames(doc, lang='en', doc_meta={'doc_id': 'mr.sun', 'doc_title': 'Our Mr. Sun'})
        return frames
    else:
        raise Exception("Unknown {} frame example".format(name))


def load_example_doc_data(names):
    cur_path = os.path.dirname(os.path.abspath(__file__))
    docs = []
    for n in names:
        doc = json.load(open(os.path.join(cur_path, '..', 'resources/{}'.format(n))))
        docs.append(doc)
    return docs

if __name__ == '__main__':
    x = load_example_doc_data(['mr-sun.json', 'technology.json'])
    print(x)
